︠1e0c7f07-74ed-45a5-ba4b-7ea95f930b18︠
from sage.ext.fast_eval import fast_float
︡123aa964-0a49-4448-a15f-5d18b71466e7︡︡
︠0b283346-6c8d-4c88-887e-3cf73bddc162i︠
%html
<p>Simple bisection root finder for a function of 3 variables.</p>

︡c2ee3c0b-65e7-44f7-a684-a21f67e3a539︡{"html": "<p>Simple bisection root finder for a function of 3 variables.</p>"}︡
︠0576a269-fcc7-4990-83c7-90935f83f932ai︠
%hide
%auto
%cython

cpdef find_root(f, double target, point_1, point_2, double epsilon=1e-4):
    """
    Given two 3d points point_1 and point_2, find a point (x,y,z) on the segment between them where abs(f(x,y,z)-target)<=epsilon.
    
    Returns (x,y,z) if the point lies on the segment between point_1 and point_2 and satisfies the above inequality, otherwise returns None.    
    
    The code assumes the function f is continuous.
    """
    cdef double s0,s1,s2
    cdef double e0,e1,e2
    cdef double new_0,new_1,new_2
    cdef double target_0 = target-epsilon
    cdef double target_1 = target+epsilon
    cdef double val
    cdef int i
    
    cdef double min = f(*point_1)
    cdef double max = f(*point_2)
    s0,s1,s2 = point_1
    e0,e1,e2 = point_2
    if min>max:
        min,max=max,min
        s0,s1,s2,e0,e1,e2 = e0,e1,e2,s0,s1,s2
    # Check to see if one of the endpoints satisfies it
    if target_0<=min and min<=target_1:
        return (s0,s1,s2)
    if target_0<=max and max<=target_0:
        return (e0,e1,e2)
    if min>target_1 or max<target_0:
        return None
    i=0
    while True:
        if i>100:
            return None
        else:
            i+=1
        # Get half-way point
        new_0 = s0+(e0-s0)/2.0
        new_1 = s1+(e1-s1)/2.0
        new_2 = s2+(e2-s2)/2.0
        val = f(new_0, new_1, new_2)
        if val<target_0:
            s0, s1, s2 = new_0, new_1, new_2
            min = val
        elif target_1<val:
            e0, e1, e2 = new_0, new_1, new_2
            max=val
        else:
            return (new_0,new_1,new_2)
︡1d63ff48-a763-41df-a942-3791e677e0e0︡︡
︠e5d1b875-481a-4e8a-a106-159f0eebb106︠
def calculate_crossing(f,target,v0,v1,vertices, cache_dict):
    """
    Calculate, for an edge (v0,v1), where f is "close" to target.  Use cache_dict to cache the values.
    """
    # Make a canonical ordering of the vertices since (v0,v1) is the same edge as (v1,v0)
    if v0>v1:
        v0,v1=v1,v0
    if (v0,v1) in cache_dict:
        return cache_dict[(v0,v1)]
    else:
        pt = find_root(f, target, vertices[v0], vertices[v1])
        cache_dict[(v0,v1)]=pt
        return pt
︡cdd9a47d-af98-4e26-bc23-e7ebd47d539c︡︡
︠fdf5234a-3788-4cd1-a66a-c58738c7c1ee︠
%time
var('x,y,z')
p=parametric_plot((x,y,9-x^2-y^2), (x,-3,3), (y,-3,3), mesh=True)
f=x^2-sin(x*y^2)+cos(z)
ff=fast_float(f, 'x', 'y','z')
p.triangulate()
vertices=p.vertex_list()

# I am still calculating the function on each vertex multiple times; that could be optimized
# vertex_values=[ff(*v) for v in vertices]

mesh=[]
for target in [0,2,..,20]:
    cache={}
    for face in [f+[f[0]] for f in p.index_faces()]:
        # Adding the [0] takes care of the edge from vertices[0] to vertices[-1]
        pts = [calculate_crossing(ff,target,face[i], face[i+1],vertices=vertices, cache_dict=cache) for i in range(len(face)-1)]
        pts = [i for i in pts if i is not None]
        mesh+=[line([pt1,pt2],thickness=3, color='black') for pt1,pt2 in Subsets(pts, 2)]
︡000b131b-bb5b-456b-ac39-39b9154d22fc︡{"stdout": "CPU time: 5.74 s,  Wall time: 5.74 s"}︡
︠5d43b329-b8c0-45b8-9da8-66417b898952︠
p+sum(mesh)
︡01a63cb9-fcb1-4d6c-83f6-27bf76ca8f64︡︡
︠f122eb1b-c6c9-48e7-bbd2-1a2e576997dd︠
(p+sum(mesh)).show(viewer='tachyon')
︡836fc340-c85c-4ab4-a66e-38293036796c︡︡
